/*
 *  prot_login.c
 *
 *  QQ Protocol. Part Login
 *
 *  Copyright (C) 2008  Huang Guan
 *
 *  2008-7-12 Created.
 *
 *  Description: This file mainly includes the functions about 
 *  Touching server, getting login token, getting password token, getting loginkey.
 *  sending login information, logging out.
 *
 */

#include <time.h>
#include <stdlib.h>
#include <string.h>
#ifdef __WIN32__
#include <winsock.h>
#include <wininet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include "qqclient.h"
#include "memory.h"
#include "debug.h"
#include "qqpacket.h"
#include "packetmgr.h"
#include "qqcrypt.h"
#include "md5.h"
#include "crc32.h"
#include "protocol.h"
#include "group.h"

#define RANDOM
#ifdef RANDOM
#endif
static int rand2()
{
	return 0;//(rand()<<16)+rand();
}

void prot_login_touch( struct qqclient* qq )
{
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_TOUCH );
	bytebuffer *buf = p->buf;
	memset( qq->server_info, 0, sizeof(qq->server_info) );
	int i;
	for( i=0; i<16; i++ )
		p->key[i] = rand();
	//15 zeros
	put_word( buf, 0 );
	put_byte( buf, 0 );
	put_int( buf, 0 );
	put_int( buf, 0 );
	put_int( buf, 0 );
	post_packet( qq, p, RANDOM_KEY );
}

void prot_login_touch_with_info( struct qqclient* qq, uchar* server_info, uchar len )
{
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_TOUCH );
	bytebuffer *buf = p->buf;
	memset( qq->server_info, 0, sizeof(qq->server_info) );
	int i;
	for( i=0; i<16; i++ )
		p->key[i] = rand();
	//15 zeros
	put_data( buf, server_info, len );
	post_packet( qq, p, RANDOM_KEY );
}

void prot_login_touch_reply( struct qqclient* qq, qqpacket* p )
{
	bytebuffer *buf = p->buf;
	ushort result = get_word( buf );
	if( result == 0x0001 ){
		memcpy( qq->server_info, buf->data, sizeof(qq->server_info) );
		uchar b1 = get_byte( buf );
		uint unit1 = get_int( buf );
		uint unit2 = get_int( buf );
		uint old_ip = qq->server_ip;
		qq->server_ip = get_int( buf );
		struct in_addr addr;
		addr.s_addr = htonl( qq->server_ip );
		DBG("redirecting to %s", inet_ntoa( addr ) );
		connect_server( qq );
		packetmgr_new_seqno( qq );
		qqpacket* next = packetmgr_new_send( qq, QQ_CMD_TOUCH );
		put_word( next->buf, result );
		put_byte( next->buf, b1 );
		put_int( next->buf, unit1 );
		put_int( next->buf, unit2 );
		put_int( next->buf, old_ip );
		post_packet( qq, next, RANDOM_KEY );
	}else if( result == 0x0000 ){
		DBG("touch ok.");
		prot_login_get_logintoken( qq );
	}
}

void prot_login_get_logintoken( struct qqclient* qq )
{
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_GET_LOGIN_TOKEN );
	packetmgr_new_seqno( qq );
	bytebuffer *buf = p->buf;
	put_byte( buf, 0 );
	post_packet( qq, p, NO_KEY );
}

void prot_login_get_logintoken_reply( struct qqclient* qq, qqpacket* p )
{
	bytebuffer *buf = p->buf;
	uchar result = get_byte( buf );
	if( result == 0 ){
		get_token( buf, &qq->login_token );
		//next
		prot_login_get_passwordtoken( qq, NULL, 0, 0 );
	}
}


void prot_login_get_passwordtoken( struct qqclient* qq, token* tok, uint code, char png_data )
{
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_GET_PASSWORD_TOKEN );
	bytebuffer *buf = p->buf;
	put_byte( buf, qq->login_token.len );
	put_data( buf, qq->login_token.data, qq->login_token.len );
	if( code )
		put_byte( buf, 4 );
	else
		put_byte( buf, 3 );
	put_byte( buf, 0 );
	put_byte( buf, 5 );
	put_int( buf, 0 );
	put_byte( buf, png_data );
	if( code ){
		put_byte( buf, 4 );
		put_int( buf, htonl( code ) );
		put_byte( buf, 0 );
		//answer token
		put_byte( buf, tok->len );
		put_data( buf, tok->data, tok->len );
	}else if( png_data ){
		put_byte( buf, 0 );
		//png token
		put_byte( buf, tok->len );
		put_data( buf, tok->data, tok->len );
	}else{
		put_byte( buf, 0 );
		put_byte( buf, 0 );
	}
	post_packet( qq, p, RANDOM_KEY );
}

void prot_login_get_passwordtoken_reply( struct qqclient* qq, qqpacket* p )
{
	token answer_token;
	token png_token;
	uchar next = 0;
	bytebuffer *buf = p->buf;
	uchar result = get_byte( buf );	//03: ok   04: need verifying
	get_byte( buf ); //00
	get_byte( buf ); //05
	uchar png_data = get_byte( buf );
//	get_int( buf );	//1205ͳ Ҫ֤ʱΪ00 00 01 23ҪʱΪȫ0
	get_byte( buf );	//00
	get_token( buf, &answer_token );
	if( png_data == 1 ){	//there's data for the png picture
		int len;
		len = get_word( buf );
		uchar* data;
		NEW( data, len );
		get_data( buf, data, len );
		get_byte( buf ); //00
		next = get_byte( buf );
		char path[PATH_LEN];
		sprintf( path, "%s/%u.png", qq->verify_dir, qq->number );
		FILE *fp;
		if( next ){
			fp = fopen( path, "wb" );
		}else{
			fp = fopen( path, "ab" );
			DBG("got png at %s", path );
		}
		if( fp ){
			fwrite( data, len, 1, fp );
			fclose( fp );
		}
		DEL( data );
		get_byte( buf ); //00
		get_token( buf, &png_token );
	}
	//parse the data we got
	if( png_data ){
		if( next ){
			prot_login_get_passwordtoken( qq, &png_token, 0, 1 );
		}else{
			qq->verify_token = answer_token;
			qqclient_set_process( qq, P_VERIFYING );
		}
	}else{
		DBG("process verify password");
		qq->token_c = answer_token;
		prot_login_verify_password( qq, &answer_token );
	}
	result = 0;
}



void prot_login_verify_password( struct qqclient* qq, token* tok )
{
	static uchar head[] = { 0x00, 0x5F, 0x00, 0x00, 0x08, 0x04, 0x01, 0xE0 };
	static uchar unknown[] = { 0xDB, 0xB9, 0xF3, 0x0B, 0xF9, 0x13, 0x87, 0xB2, 
		0xE6, 0x20, 0x43, 0xBE, 0x53, 0xCA, 0x65, 0x03 };
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_VERIFY_PASSWORD );
	int i;
	for( i=0; i<sizeof(unknown); i++ )
		unknown[i] = rand2();
	bytebuffer *buf = p->buf;
	put_data( buf, head, sizeof(head) );
	put_byte( buf, tok->len );
	put_data( buf, tok->data, tok->len );
	put_byte( buf, 0 );
	//prepare the password verification encrypted data
	uchar data[20], encrypted[48];
	memcpy( data, qq->md5_pass1, 16 );
	data[16] = data[17] = 0;
	*(ushort*)&data[18] = rand()&0xffff;
	int out_len = 48;
	qqencrypt( data, 20, qq->md5_pass2, encrypted, &out_len );
	put_byte( buf, out_len );
	put_data( buf, encrypted, out_len );
	//
	put_byte( buf, 0 );
	put_byte( buf, sizeof(unknown)+sizeof(uint) );
	put_data( buf, unknown, sizeof(unknown) );
	put_int( buf, crc32( unknown, sizeof(unknown) ) );
	put_byte( buf, 0 );
	put_byte( buf, 3 );	//len of the remaining
	put_byte( buf, 0 );
	put_byte( buf, data[1] );
	put_byte( buf, data[2] );
	post_packet( qq, p, RANDOM_KEY );
}

void prot_login_verify_password_reply( struct qqclient* qq, qqpacket* p )
{
	uchar login_key[16];
	token data_v;
	bytebuffer *buf = p->buf;
	get_word( buf );	//00 97  or  00 6d()  or 00 c6()
	int code = get_byte( buf );
	switch( code ){
	case 0x00:
		buf->pos = 76;
		get_token( buf, &data_v );
		get_data( buf, login_key, 16 );
		//success
		prot_login_send_info( qq, login_key, &data_v );
		return;
	case 0x33:
	case 0x51:	//
		qqclient_set_process( qq, P_DENIED );
		DBG("Denied.");
		break;
	case 0xBF:
		DBG("No this number.");
	case 0x34:
		DBG("Wrong password.");
		qqclient_set_process( qq, P_WRONGPASS );
		break;
	default:
		hex_dump( buf->data, buf->len );
		break;
	}
	buf->pos = 11;
	uchar len = get_word( buf );
	char msg[256];
	get_data( buf, (uchar*)msg, len );
	msg[len] = 0;
	DBG( msg );
	//failed
}



void prot_login_send_info( struct qqclient* qq, uchar* key, token* data_v )
{
//	static uchar qq_hash[] = { 0x23, 0xFC, 0xD2, 0xEB, 0x9E, 0x6D, 0xE0, 
//		0x73, 0x11, 0xA1, 0x16, 0xEB, 0xF1, 0x05, 0x0C, 0xA7 };
	static uchar qq_hash[] = { 0xD2, 0x41, 0x75, 0x12, 0xC2, 0x86, 0x57, 
		0x10, 0x78, 0x57, 0xDC, 0x24, 0x8C, 0xAA, 0x8F, 0x4E };
	static uchar unknown2[] = { 0x94,0x0B,0x73,0x7A,0xA2,0x51,0xF0,0x4B,
		0x95,0x2F,0xC6,0x0A,0x5B,0xF6,0x76,0x52 };
	static uchar unknown3[] = { 0x00,0x00,0x00,0x06,0x00,0x00,0x00,0x00,
		0x00,0x00,0x01,0x40,0x01 };
	static uchar unknown3_5[]= { 0x2D,
		0x49,0x15,0x55,0x78,0xFC,0xF3,0xD4,0x53,0x55,0x60,0x9C,0x37,
		0x9F,0xE9,0x59 };
	static uchar unknown4[] = { 0x3B,
		0xCE,0x43,0xF1,0x8B,0xA4,0x2B,0xB5,0xB3,0x51,0x57,0xF7,0x06,
		0x4B,0x18,0xFC };
	static uchar unknown5[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00 };
	int i;
	//rand2 unknowns
#ifdef RANDOM
	for( i=0; i<sizeof(unknown2); i++ )
		unknown2[i] = rand2();
	for( i=0; i<sizeof(unknown3_5); i++ )
		unknown3_5[i] = rand2();
	for( i=0; i<sizeof(unknown4); i++ )
		unknown4[i] = rand2();
#endif
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_LOGIN );
	bytebuffer *buf = p->buf;
	memcpy( p->key, key, 16 );
	put_word( buf, 0 );	//unknown
	put_byte( buf, 0 );	//separator
	
	//prepare the password verification encrypted data
	uchar data[20], encrypted[48];
	memcpy( data, qq->md5_pass1, 16 );
	data[16] = data[17] = 0;
	*(ushort*)&data[18] = rand()&0xffff;
	int out_len = 48;
	qqencrypt( data, 20, qq->md5_pass2, encrypted, &out_len );
	put_byte( buf, out_len );
	put_data( buf, encrypted, out_len );
	
	//մ
	out_len = 48;
	qqencrypt( (uchar*)"", 0, qq->md5_pass2, encrypted, &out_len );
	put_data( buf, encrypted, out_len );
	for( i=0; i<19; i++ )
		put_byte( buf, 0 );
	put_data( buf, qq_hash, sizeof( qq_hash ) );
	//local index
	//note: encrypted is not encrypted now
	memcpy( encrypted+out_len, qq_hash, sizeof( qq_hash ) );
	out_len += sizeof( qq_hash );	//out_len must be 32
	
	uchar index = rand()%3;	// always 1 is ok too.
	for( i=0; i<out_len; i++ )
		index ^= encrypted[i];
	put_byte( buf, index );
	put_byte( buf, qq->mode );
	put_data( buf, unknown5, sizeof( unknown5 ) );
	put_data( buf, qq->server_info, 15 );
	put_data( buf, unknown2, sizeof( unknown2 ) );
	put_byte( buf, qq->token_c.len );
	put_data( buf, qq->token_c.data, qq->token_c.len );
	put_data( buf, unknown3, sizeof( unknown3 ) );
//	0x1B,0x02,0x84,0x50
#ifdef RANDOM
	put_int( buf, rand2() );
#else
	put_int( buf, 0x1B028450 );
#endif
	put_byte( buf, 00 );
	put_byte( buf, sizeof( unknown3_5 ) );
	for( i=0; i<sizeof(unknown3_5); i++ )
		put_byte( buf, unknown3_5[i] );
	put_data( buf, unknown5, sizeof( unknown5 ) );
	put_data( buf, qq->server_info, 15 );
	put_byte( buf, 0x02 );
//	0x68,0xE8,0x07,0x83
#ifdef RANDOM
	put_int( buf, rand2() );
#else
	put_int( buf, 0x68E80783 );
#endif
	put_byte( buf, 00 );
	put_byte( buf, sizeof( unknown4 ) );
	for( i=0; i<sizeof(unknown4); i++ )
		put_byte( buf, unknown4[i] );
	for( i=0; i<249; i++ )
		put_byte( buf, 0 );
	qq->data_v = *data_v;
	post_packet( qq, p, RANDOM_KEY );
}

void prot_login_send_info_reply( struct qqclient* qq, qqpacket* p )
{
	bytebuffer *buf = p->buf;
//	hex_dump( buf->data, buf->len );
	uchar result = get_byte( buf );
	if( result != 0 )
	{
		DBG("login result = %d", result );
		qqclient_set_process( qq, P_ERROR );
		return;
	}
	get_data( buf, qq->session_key, 16 );
	DBG("session key: " );
	hex_dump( qq->session_key, 16 );
	if( qq->number != get_int( buf ) ){
		DBG("qq->number is wrong?");
	}
	qq->client_ip = get_int( buf );
	qq->client_port = get_word( buf );
	qq->local_ip = get_int( buf );
	qq->local_port = get_word( buf );
	qq->login_time = get_int( buf );
	get_byte( buf );	//03
	get_byte( buf );	//mode
	//prepare IM key
	uchar data[20];
	*(uint*)data = htonl( qq->number );
	memcpy( data+4, qq->session_key, 16 );
	//md5
	md5_state_t mst;
	md5_init( &mst );
	md5_append( &mst, (md5_byte_t*)data, 20 );
	md5_finish( &mst, (md5_byte_t*)qq->im_key );
	//
	buf->pos = 135;
	qq->last_login_time = get_int( buf );
	time_t t;
	t = CN_TIME( qq->last_login_time );
	DBG("last login time: %s", ctime( &t ) );
	qqclient_set_process( qq, P_LOGIN );
	//prepare keys
	static uchar key_types[] = { 0x03, 0x04, 0x06, 0x07, 
		0x08, 0x0a, 0x0b, 0x0c, 0x0e };
	int i;
	for( i=0; i<sizeof(key_types); i++ )
		prot_user_get_key( qq, key_types[i] );
	//get information
	prot_user_get_notice( qq, 0 );
	prot_user_get_notice( qq, 1 );
	prot_user_get_info( qq, qq->number );
	buddy_update_list( qq );
	prot_user_check_ip( qq );	//cheat
	prot_user_change_status( qq );
	prot_misc_get_ad( qq, 7 );
	prot_misc_unknown_b8( qq );	//cheat
	prot_misc_unknown_c6( qq );	//cheat
	prot_misc_unknown_2e( qq );	//cheat
	group_update_list( qq );
	prot_misc_version( qq );	//cheat
	prot_user_get_level( qq, 8 );
	qq->online_clock = 0;
}


void prot_login_logout( struct qqclient* qq )
{
	qqpacket* p = packetmgr_new_send( qq, QQ_CMD_LOGOUT );
	bytebuffer *buf = p->buf;
	char data[32];
	uchar md5[16];
	sprintf( data, "%s%u", qq->self->nickname, qq->number );
	md5_state_t mst;
	md5_init( &mst );
	md5_append( &mst, (md5_byte_t*)data, strlen( data ) );
	md5_finish( &mst, (md5_byte_t*)md5 );
	put_data( buf, md5, 16 );
	p->need_ack = 0;
	post_packet( qq, p, SESSION_KEY );
}
